#!/usr/bin/env python

from types import *
from math import atan2
from VTKHelperFunctions import *
from masks import *
from ephem import *
from basics import *

class astro:
    def __init__(self,n,suffix="ast"):
       self.image = n
       self.name = replace(n,".fits",".soln")
       self.init = readsoln(self.name)
       self.catalog = replace(self.image,".fits",".%s" % (suffix))
       self.rac, self.decc = self.init["sky"]
       self.ra = formatSexagesimal(radhr(self.rac),4,3600000)
       self.dec = formatSexagesimal(raddeg(self.decc),4,3600000)
       self.tran = self.init["c"]
       self.nterms = len(self.tran)
       self.tranx = take(self.tran,range(0,self.nterms,2))
       self.trany = take(self.tran,range(1,self.nterms,2))
       self.scale = sqrt(self.tran[2]**2 + self.tran[3]**2)
       self.xc, self.yc = self.init["center"]
    def write(self,verb=0):
       if verb: print "Writing",self.name
       f=open(self.name,"w")
       f.write("%3d 0 %9.3f %9.3f 0.0\n" % (len(self.tran),self.xc,self.yc))
       f.write("%13s %13s\n" % (self.ra,self.dec))
       [f.write("%17.9g\n" % (k)) for k in self.tran]
       f.close()
    def sex(self,sexdir="./sex/",thresh=5.0,saturate=30000,seeing=1.0):
       self.sexcom = ["sex",self.image]
       self.sexcom += ["-c "+sexdir+"default.sex"]
       self.sexcom += ["-PARAMETERS_NAME "+sexdir+"default.param"]
       self.sexcom += ["-DETECT_MINAREA 5"]
       self.sexcom += ["-DETECT_THRESH "+str(thresh)]
       self.sexcom += ["-ANALYSIS_THRESH "+str(thresh)]
       self.sexcom += ["-PIXEL_SCALE "+str(self.scale)]
       self.sexcom += ["-SEEING_FWHM "+str(seeing/self.scale)]
       self.sexcom += ["-CATALOG_NAME",self.catalog]
#        self.sex += ["-FILTER_NAME "+sexdir+"default.conv"]
       self.sexcom += ["-FILTER_NAME "+sexdir+"gauss_3.0_5x5.conv"]
       self.sexcom += ["-STARNNW_NAME "+sexdir+"default.nnw"]
       self.sexcom += ["-SATUR_LEVEL "+str(saturate)]
       self.sexcom = join(self.sexcom)
       verbex(self.sexcom)
    def readcat(self,refsim=None,nuse=None,clean=0):
       self.nuse = nuse
       self.dat = readtable(self.catalog,icols=[0,7],start=9)
       self.imx = self.dat[1]; self.imy = self.dat[2]
       self.gra,self.gdec = xy2rd(self.name,self.imx,self.imy,text=0)
       self.nx, self.ny = normalcoord(self.gra,self.gdec,self.rac,self.decc,conv=3600*180/pi)
       self.dat[1], self.dat[2] = self.nx, self.ny
       self.cra, self.cdec = xy2rd(self.name,array([0.0,2*self.xc]),array([0.0,2*self.yc]),text=0)
       self.cx, self.cy = normalcoord(self.cra,self.cdec,self.rac,self.decc,conv=3600*180/pi)
       self.catdat = self.dat
       self.catmatch = matching(self.catdat,nuse=self.nuse,cull=clean,errlim=0.2)
       if refsim:
          self.refsim = refsim
          refdat = readsim(refsim)[0]
          refra = array([hrrad(scanSexagesimal(r)) for r in refdat[3]])
          refdec = array([degrad(scanSexagesimal(d)) for d in refdat[4]])
          refnx, refny = normalcoord(refra,refdec,self.rac,self.decc,conv=3600*180/pi)
          keep = between(refnx,min(self.cx),max(self.cx)) * \
                  between(refny,min(self.cy),max(self.cy))
          self.refid = [refdat[0][i] for i in nonzero(keep)]
          self.refmag  = array([refdat[2][i] for i in nonzero(keep)])
          self.refnx  = array([refnx[i] for i in nonzero(keep)])
          self.refny  = array([refny[i] for i in nonzero(keep)])
          self.refdat = [self.refid,self.refnx,self.refny,self.refmag]
          self.refmatch = matching(self.refdat,sky=1,cull=0,nuse=self.nuse)
    def matchup(self,verb=0,mtol=1,ttol=1,nterms=1):
       self.fixsol,self.matches = self.catmatch.match(1,self.refmatch,verb=verb,im=self.image,mtol=mtol,ttol=ttol)
#       self.rotsol = rotsol(self.fixsol)
#       self.fix()
       ix = take(self.imx,self.matches[0])
       iy = take(self.imy,self.matches[0])
       rx = (pi/180)*take(self.refnx,self.matches[1])/3600
       ry = (pi/180)*take(self.refny,self.matches[1])/3600
       for j in range(1):
          b = abasis(nterms,ix-self.xc,iy-self.yc)
          y = concatenate([rx,ry])
          self.tran = svdfit(b,y)
#          if j: self.tran = planefit(b,y,self.tran,minfunc=mad)[0]
          f = add.reduce(self.tran[NewAxis,::]*b,1)
          fx = f[:len(ix)]; fy = f[len(ix):]
       if verb or 1: print multiply(bwt(fx-rx),(3600*180/pi)),
       if verb or 1: print multiply(bwt(fy-ry),(3600*180/pi)),len(ix)
    def touchup(self,verb=0,mtol=1.0,clean=0,nord=0):
        self.catmatch = matching(self.catdat,sky=0,cull=clean,verb=0,nuse=None)
        self.refmatch = matching(self.refdat,sky=1,cull=0,verb=0,nuse=None)
        self.fixsol = self.refmatch.xfm(self,verb=0,nord=0,mtol=mtol,arcsec=1)
        self.rotsol = rotsol(self.fixsol)
        self.fix(verb=verb)
    def fix(self,verb=0):
        if verb: print "Old:",self.tran
        self.xsh = degrad(self.rotsol.xsh/3600)
        self.ysh = degrad(self.rotsol.ysh/3600)
        self.cos = self.rotsol.cos
        self.sin = self.rotsol.sin
        self.rescale = self.rotsol.rescale
        self.tranx, self.trany = self.tranx*self.cos - self.trany*self.sin,\
                 self.tranx*self.sin + self.trany*self.cos
        self.tranx[0], self.trany[0] = self.tranx[0] - self.xsh, self.trany[0] - self.ysh
        self.tran = concatenate([[self.tranx[j],self.trany[j]] for j in range(self.nterms/2)])
        if verb: print "New:",self.tran

class rotsol:
   def __init__(self,fixsol,fac=1.0):
      self.rot = (pi/2 - apply(atan2,fixsol[2:]))
      self.xsh,self.ysh = fixsol[:2]
      self.cos = cos(self.rot)
      self.sin = sin(self.rot)
      self.rescale = 1.0/sum(fixsol[2:]**2)
      print "Shifts, dPA, Scale = (%.2f,%.2f) (arcsec), %.3f (deg), %3.3f " % (self.xsh,self.ysh,raddeg(self.rot),self.rescale)

class matching:
   def __init__(self,d,nuse=50,errlim=0.1,cull=1,sky=0,verb=0):
       if sky:
          n, x, y, m= d
          e = zeros(x.shape)
          fl = zeros(x.shape)
          u = arange(len(n))
          n = arange(len(n))
          un = n
          ux = x
          uy = y
          um = m
          ue = e
          uf = 1*e
       else:
          n, x, y, m, e, ma, ea, fl = d[:8]
          n = arange(len(x))
          u = arange(len(n))

       if cull:
          u = less(ea,errlim)*between(ma-m,1.0,4.0)
#          u = u*not_equal(fl,4)*less(fl,30)
          u = u*less(fl,30)
          u = nonzero(u)
          n = take(n,u)
          x,y,m,e,ma,ea,fl = take([x,y,m,e,ma,ea,fl],u,1)
          un = n; ux,uy,um,ue = x,y,m,e
#          un = take(n,u); ux,uy,um,ue = take([x,y,m,e],u,1)
          print "Culling",len(un),
          if nuse:
              a = argsort(m)[:nuse]
              n = take(n,a)
              x,y,m,e,ma,ea,fl = take([x,y,m,e,ma,ea,fl],a,1)
              u = arange(len(x))
              un = n; ux,uy,um,ue = x,y,m,e
              print "and using",len(un)
          else: print
       else:
          a = argsort(m)
          if nuse: a = a[:nuse]
          n = take(n,a)
          x,y,m,e, = take([x,y,m,e],a,1)
          un = n; ux,uy,um,ue = x,y,m,e
          u = arange(len(n))
       if nuse:             
           if verb: print "There are",len(n),"reference objects."

           id = []
           n1 = []; x1 = []; y1 = []; m1 = []; e1 = []; s1 = []; t1=[]
           for j in u:
              n2 = []; x2 = []; y2 = []; m2 = []; e2 = []; s2 = []; t2=[]
              for k in u:
                 dx = x[j] - x[k]
                 dy = y[j] - y[k]
                 ds = sqrt(dx**2+dy**2)
                 dt = raddeg(atan2(dy,dx)-pi/2)
                 n2.append( n[k]); x2.append( x[k]); y2.append( y[k])
                 m2.append( m[k]); e2.append( e[k])
                 s2.append( ds); t2.append( dt)
              n2 = asarray(n2,Int32)
              x2,y2,m2,e2,s2,t2 = asarray([x2,y2,m2,e2,s2,t2])
              o2 = argsort(s2)
              n2 = take(n2,o2)
              x2,y2,m2,e2,s2,t2 = take([x2,y2,m2,e2,s2,t2],o2,1)
              id.append(n[j])
              n1.append(n2); x1.append(x2); y1.append(y2)
              m1.append(m2); e1.append(e2)
              s1.append(s2); t1.append(t2)
           id = asarray(id,Int32)
           n1 = asarray(n1,Int32)
           x1,y1,m1,e1,s1,t1 = asarray([x1,y1,m1,e1,s1,t1])
           self.id = id; self.n1 = n1; self.x1 = x1; self.y1 = y1
           self.m1 = m1; self.e1 = e1; self.s1 = s1; self.t1 = t1
       self.n = un; self.x = ux; self.y = uy
       self.m = um; self.e = ue

   def match(self,Jim,other,verb=0,im=None,mtol=1.0,ttol=1.0,simple=0):
       sys.stdout.flush()
       print "Entering matching algorithm:",verb
       J1,J2=len(self.n1),len(other.n1)
       b = []; n = []; x = []; y = []
       for j1 in range(J1):
          if len(n) == J2: break
          print j1,
          if j1 and not j1%25: print
          sys.stdout.flush()
          possx = []
          possy = []
          possn = []
          possb = []
          possm = []
          for j2 in range(J2):
            if j2 not in b:
               ts = (self.s1[j1][::,NewAxis]-other.s1[j2][NewAxis,::])
               ss = less(abs(ts),mtol)
               ss[0,0] = 0
               if verb == 3:
                  print
                  print j1,j2
                  for temp in ts: print temp.astype(Int32)
                  for temp in ss: print temp.astype(Int32)
#               if sum(ss.flat) >= min([J1,J2])/4:
               if sum(ss.flat) > 0:
                   tt = self.t1[j1][::,NewAxis]-other.t1[j2][NewAxis,::]
#                   tt = sin((pi/180)*tt)
                   st = less(abs(tt),ttol)
                   st[0,0] = 0
                   mat = (ss*st)
                   tmat = sum(mat.flat)
                   if verb == 3:
                      for temp in tt: print temp.astype(Int32)
                      for temp in st: print temp.astype(Int32)
                      print
                      for temp in mat: print temp
                      print tmat
                   if tmat > 3:
#                   if tmat >= min([J1,J2])/4:
                      if verb > 1:
                          print j1,j2,tmat
                          spaces = compress(mat.flat,ts.flat)
                          thetas = compress(mat.flat,tt.flat)
                          print spaces
                          print thetas
                          print
#                      x.append(self.x[j1])
#                      y.append(self.y[j1])
#                      n.append(self.n[j1])
#                      b.append(j2)
                      possx.append(self.x[j1])
                      possy.append(self.y[j1])
                      possn.append(self.n[j1])
                      possb.append(j2)
                      possm.append(tmat)
                      break
          if len(possm):
            if verb > 1:
               print j1, possb
            use = argmax(possm)
            x.append(possx[use])
            y.append(possy[use])
            b.append(possb[use])
            n.append(possn[use])
       x,y = asarray([x,y])
       print
       noth = take(other.n,b)
       xo = take(other.x,b)
       yo = take(other.y,b)
       if verb: print xo-x
       if verb: print yo-y
       bx,sx = bwt(x-xo)
       by,sy = bwt(y-yo)
       if verb: len(xo)
       if verb: print bx,sx
       if verb: print by,sy
       usex = between(x-xo,bx-4*sx,bx+4*sy)
       usey = between(y-yo,by-4*sy,by+4*sy)
       if verb: print usex*usey
       xo,yo,x,y = compress(usex*usey,[xo,yo,x,y],1)
       noth,n = compress(usex*usey,[noth,n],1)
       if verb: print xo-x
       if verb: print yo-y
       print "Matched:",len(xo)
       if simple:
          bx = bwt(x-xo)
          by = bwt(y-yo)
          rt = array([(atan2(y[j]-mean(y),x[j]-mean(x))-atan2(yo[j]-mean(yo),xo[j]-mean(xo))) for j in range(len(x))])
          bt = bwt(rt)
          return bx,by,bt
       else:
          basis = abasis(0,xo,yo,rot=1)
          dat = concatenate([x,y])
          sol = svdfit(basis,dat)
          print sol
          f = add.reduce(sol[NewAxis,::]*basis,1)
          fx = f[:len(x)]; fy = f[len(x):]
          return sol,[n,noth]

   def xfm(self,wirc,tran=[0,0,1,0],verb=0,mtol=3.0,nord=1,arcsec=0):
      other = wirc.catmatch
      if tran: tran = array(tran)
      else: tran = wirc.tran
      if arcsec:
         fac = 1.0; xc = 0.0; yc = 0.0
         aix = self.x; aiy = self.y
         arx = other.x; ary = other.y
         print len(aix), len(arx)
      else:
         fac = (3600*180/pi)
         xc = wirc.xc; yc = wirc.yc
         aix = wirc.imx - xc
         aiy = wirc.imy - yc
         arx = other.x; ary = other.y
      nords = [0,1,2,3,4,5,6]
      ltrans = [4,6,12,20,30,42,56]
      for j in range(3):
          rx = []; ry = []; ix = []; iy = []; fx = []; fy = []
          b = abasis(nords[ltrans.index(len(tran))],aix,aiy,rot=1)
          f = add.reduce(tran[NewAxis,::]*b,1)*fac
          afx = f[:len(aix)]; afy = f[len(aix):]
          for k in range(len(aix)):
             dx = power(afx[k]-arx,2)
             dy = power(afy[k]-ary,2)
             ds = sqrt(dx+dy)
             l = argmin(ds)
             if ds[l] < mtol:
                ix += [aix[k]]; iy += [aiy[k]]
                fx += [afx[k]]; fy += [afy[k]]
                rx += [arx[l]]; ry += [ary[l]]
          ix,iy,rx,ry,fx,fy = asarray([ix,iy,rx,ry,fx,fy])
          mx,dx = bwt(fx-rx)
          my,dy = bwt(fy-ry)
          if verb: print mx,dx,mx,dy,len(fx)
          sel = between(fx-rx,mx-4*dx,mx+4*dx)*between(fy-ry,my-4*dy,my+4*dy)
          ix = compress(sel,ix,0)
          iy = compress(sel,iy,0)
          rx = compress(sel,rx,0)
          ry = compress(sel,ry,0)
          b = abasis(nord,ix,iy,rot=1)
          y = concatenate([rx,ry])
          tran = svdfit(b,y)
#          tran = planefit(b,y,tran,minfunc=mad)[0]
          tran = tran/fac
          f = add.reduce(tran[NewAxis,::]*b,1)*fac
      fx = f[:len(ix)]; fy = f[len(ix):]
      mx,dx = bwt(fx-rx)
      my,dy = bwt(fy-ry)
      print "Sigmas (x,y)=(%.4f,%.4f) [%d]" % (dx,dy,len(fx))
      return tran

