
from SimpleQueue import *
from basics import *
from MLab import *
from FITS import *
from Multipack import *
from VTKHelperFunctions import *
from scipy import optimize

def GetImage(image,dobias=0):
    F = FITS(image)
    noverscn = F["noverscn"]
    nbiaslns = F["nbiaslns"]
    data = F.data().astype(Float32)
    F.close()
    if dobias:
       print "Computing column bias."
       colbias = median(transpose(data[::,data.shape[1]-noverscn: data.shape[1]]))
       print "Subtracting column bias."
       data = subtract(data,colbias[::,NewAxis].astype(Float32),data)
       print "Computing row bias."
       rowbias = median(data[data.shape[0]-nbiaslns: data.shape[0]])
       print "Subtracting row bias."
       data = subtract(data,rowbias[NewAxis,::].astype(Float32),data)
       data = data[:-nbiaslns,:-noverscn].astype(Float32)
    else:
       if (data.shape[0] % 1024): data = data[:-nbiaslns,::]
       if (data.shape[1] % 1024): data = data[::,:-noverscn]
       data = data.astype(Float32)
    return data

def RingKernel(Data,ring,span):
    iring = int(ring+span/2+2)
    Y,X = indices((2*iring+1,2*iring+1),Float64)
    Xc = Yc = iring
    X, Y = X-Xc, Y-Yc
    R = sqrt(Y*Y+X*X)
#    sKernel = getexp(-0.5*power((R-ring)/(span/2.35),2))
#    sKernel = (sKernel*between(R,ring-span/2.,ring+span/2.)).astype(Float32)
    sKernel = between(R,ring-span/2.,ring+span/2.).astype(Float32)
    sKernel = sKernel / sum(sKernel.flat)
#    qdump("k.fits",sKernel)
    Kernel = zeros(Data.shape,Float32)
    Kernel[ Data.shape[0]/2-Yc: Data.shape[0]/2+Yc+1, Data.shape[1]/2-Yc: Data.shape[1]/2+Yc+1] = sKernel.astype(Float32)
    return Kernel

def GetPositions(Convolved,dist=20,M=4e3,th=300,maxval=60e3,clean=1):
    print "Invoking Non-Maximum Suppression."
    Cl = VTKNonMaxSuppress(Convolved,numret=1,usegrad=0)
    print "Clipping between %.1e and %.1e." % (M+th,maxval)
    l = between(Cl,M+th,maxval).astype("b")
    print "Removing \"islands\" with areas more than two pixels."
    l = VTKSubtract(l, VTKIslandRemoval(l,1,0,2,numret=1))
    Cl = (Convolved*l).astype(Float32)
#    qdump("l.fits",Cl)
    u = nonzero(l.flat)
    print "Culling out %d candidate objects." % (len(u))
    z = take(Cl.flat,u) - M
    x = fmod(u,Cl.shape[1])
    y = (u-x)/Cl.shape[1]
    if clean:
       print "Adding up neighbours."
       dx = (x[::,NewAxis] - x[NewAxis,::]).astype(Float32)
       dy = (y[::,NewAxis] - y[NewAxis,::]).astype(Float32)
       dz = greater(z[::,NewAxis], z[NewAxis,::])
       ds = VTKSqrt(VTKMultiply(dx,dx)+VTKMultiply(dy,dy))
       p = between(ds,0,dist)
       g = add.reduce(p,1)
       h = add.reduce(p*dz,1)
       k = 1.0*greater_equal(h,g)*less(g,1)
       print "Reducing %d candidates to %d objects." % (len(x),sum(k))
       x,y,z = compress(k,[x,y,z])
#    for j in range(len(g)):
#       print "%10.f %10.f %10.3f %5d %10.3f %4d" % (x[j], y[j], z[j], g[j], h[j], k[j])
    i = 1+arange(x.shape[0])
    return i,x,y,z

def Convolve(data,kernel):
    ff = FFT.real_fft2d(data)
    fg = FFT.real_fft2d(kernel)
    ff = multiply(ff,fg,ff)
    result = FFT.inverse_real_fft2d(ff)
    result=shiftim(result,result.shape[1]/2,result.shape[0]/2).astype(data.typecode())
    return result

def Surface(data,wid=20,maxval=64e3,th=2,sh=4):
    wid = int(wid)
    ush = int(log(wid)/log(2)) - sh + 2
    ush = max([ush,2])
    sdata = VTKShrinkN(data,sh,numret=1,usemed=1)
    m = median(sdata.flat)
    s = 1.49*median(VTKAbs(sdata.flat-m))
    good = between(sdata,m/10,m+th*s).astype(Float32)
    large = VTKDivide(VTKLevel(good*sdata,ush), VTKLevel(good,ush))
    good = less(abs(sdata-large),th)
    large = VTKDivide(VTKLevel(good*sdata,1+ush), VTKLevel(good,1+ush))
    norm = median(compress(good.flat,large.flat))
    large = large/norm
    large = large*greater(large,0.5)
    large = large.astype(Float32)
    large = VTKGrowN(large,sh,origsize=data.shape,interp=1,numret=1,mirror=1,xshift=-(sh-1)/2.,yshift=-(sh-1)/2.)
    return large,m,s

def FitHistoCounts(D,dn,verb=0):
   N = D.shape[0]
   nbins = int(sqrt(N))
   S = arange(1,N,dn,Float32)
   S0 = S.astype(Int32)
   S1 = (S+0.5).astype(Int32)
   H0 = take(D,S0,0)
   H1 = take(D,S1,0)
   H = (H0+H1)/2.0
   if verb > 1: print "doing dH"
   dH = MLab.diff(H)
   if verb > 1: print "doing dS"
   dS = MLab.diff(S)
   if verb > 1: print "concatenating dH"
   dH = concatenate([dH,[dH[-1]]])
   if verb > 1: print "concatenating dS"
   dS = concatenate([dS,[dS[-1]]])
   if verb > 1: print "dividing"
   dS = divz(dS,dH)
   xc = median(H)
   xr = 1.49*mad(H)
   uH = compress(between(H-xc,-5*xr,10*xr),H)
   xc = median(uH)
   xr = 1.49*mad(uH)
   if verb > 1: print xc,xr
   guess = [xc,log10(xr)]
   if verb > 1: print "Fitting:",guess
   H,dS = compress(greater(dS,0),[H,dS],1)
   gsol,gmsg = leastsq(fitgaussian,guess,args=(H,dS,1,verb),xtol=1e-6,ftol=1e-6,epsfcn=1e-7,factor=10.0,maxfev=5000)
   gsol[1] = 10**gsol[1]
   if verb: print " %9.1f %9.1f" % tuple(gsol)
   return gsol

def GaussianBG(data,dn,iter=5,verb=0):
   bg = FitHistoCounts(data.flat,dn,verb=verb)
   return bg

def MakeDonut(parameters,X,Y,Z,W,bg=1):
    x1,y1,a1,b1,t1,x2,y2,a2,b2,t2,lseeing = parameters
    a1,b1,a2,b2 = power(10, [a1,b1,a2,b2])
    seeing = pow(10,lseeing)
    p1, p2 = arctan(t1), arctan(t2)
#    print "%6.2f %6.2f %6.2f %6.2f %6.1f" % (x1,y1,a1,b1,p1*180/pi),
#    print "%6.2f %6.2f %6.2f %6.2f %6.1f" % (x2,y2,a2,b2,p2*180/pi),
#    print "%6.3f" % ( seeing)
    X1,Y1 = (X-x1)*cos(p1) + (Y-y1)*sin(p1), -(X-x1)*sin(p1) + (Y-y1)*cos(p1)
    X2,Y2 = (X-x2)*cos(p2) + (Y-y2)*sin(p2), -(X-x2)*sin(p2) + (Y-y2)*cos(p2)
    R1 = (X1/a1)**2 + (Y1/b1)**2
    R2 = (X2/a2)**2 + (Y2/b2)**2
    E1 = sqrt((1.0/a1)**2 + (1.0/b1)**2)
    E2 = sqrt((1.0/a2)**2 + (1.0/b2)**2)
#    M1 = getexp(-0.5*power((R1-1.0)/E1,2)*less(R1,1.0))
#    M2 = getexp(-0.5*power((R2-1.0)/E2,2)*greater(R2,1.0))
    M1 = greater(R1,1.0)
    M2 = less(R2,1.0)
    M = M1*M2

    o = 2
    x,y,z,w,m = compress(W.flat*greater(R2.flat,1.25),[X2.flat,Y2.flat,Z.flat,W.flat,M.flat])
    bases = transpose([power(x,i)*power(y,j) for i in range(o) for j in range(o-i)])
    zsol = svdfit(bases*w[::,NewAxis],z*w)
    bases = [power(X2.flat,i)*power(Y2.flat,j) for i in range(o) for j in range(o-i)]
    B = reshape(dot(zsol,bases),M.shape)

    o = 2
    x,y,z,w,m = compress(W.flat*equal(M.flat,1.0),[X2.flat,Y2.flat,Z.flat-B.flat,W.flat,M.flat])
    zm = divz(z,m)
    bases = transpose([power(x,i)*power(y,j) for i in range(o) for j in range(o-i)])
    zsol = svdfit(bases*w[::,NewAxis],zm*w)

    def see(zsol):
       bases = [power(X2.flat,i)*power(Y2.flat,j) for i in range(o) for j in range(o-i)]
       O = reshape(dot(zsol,bases),M.shape) + B
       Q = (Z - VTKGauss(O,seeing,numret=1))*W
#       U = less(R2,1.21)*greater(R1,0.81)
       U = M
       return Q.flat*U.flat

    zsol = leastsq(see,zsol,maxfev=1000,full_output=0,\
        factor=0.01,epsfcn=1e-2,gtol=1e-8,ftol=1e-8,xtol=1e-8)[0]
    bases = [power(X2.flat,i)*power(Y2.flat,j) for i in range(o) for j in range(o-i)]
    O = reshape(dot(zsol,bases),M.shape)

    F = M*O + B
    F = VTKGauss(F,seeing,numret=1)
    if not bg: F = F - B
    return F

def ResDonut(parameters,X,Y,Z,W,fmin=0):
    F = MakeDonut(parameters,X,Y,Z,W)
#    qdump("c.fits",F)
#    exit()
    res = F-Z
    val = W*res
    if fmin: s = sum(power(val.flat,2))
    else: s = val.flat
#    print sum(power(val.flat,2))
    return s

def FitDonut(Z,M,W,ring,span):
    mx,my,f,r,theta,a,b = M
    print mx,my,f,r,theta,a,b
    t = tan(theta*pi/180)
    seeing = 0.85
    x1,y1,a1,b1,t1,x2,y2,a2,b2,t2 = mx,my,max([2,ring-span/2]),max([2,(ring-span/2)*b/a]),t,mx,my,ring+span/2,(ring+span/2)*b/a,t
    donut = array([x1,y1,log10(a1),log10(b1),t1,x2,y2,log10(a2),log10(b2),t2,log10(seeing)])
    Y,X = indices(Z.shape)
#    print donut
#    donut = optimize.fmin(ResDonut,donut,args=(X,Y,Z,W,1))
    print donut
    sol = leastsq(ResDonut,donut,args=(X,Y,Z,W),maxfev=5000,full_output=1,\
        factor=0.001,epsfcn=1e-2,gtol=1e-8,ftol=1e-8,xtol=1e-8)
    donut = sol[0]
#    print donut
#    sol = leastsq(ResDonut,donut,args=(X,Y,Z,W),maxfev=5000,full_output=1,\
#        factor=0.001,epsfcn=1e-4,gtol=1e-8,ftol=1e-8,xtol=1e-8)
#    donut = sol[0]
    print donut
    c = MakeDonut(donut,X,Y,Z,W,bg=0)
    return donut,c

def Characterize(objid,x,y,z,Data,Show,Mask,ring=10,span=5,maxval=50e3,seeing=0,output=None,verb=0,inter=0,nbox=2):
    inx = nbox*(ring+span) < x < Data.shape[1]-nbox*(ring+span)
    iny = nbox*(ring+span) < y < Data.shape[0]-nbox*(ring+span)
    if not inx or not iny: return inter
    wid = int(nbox*(ring+span)+1)
    x0,x1 = clip([x-wid,x+wid],0,Data.shape[1]).astype(Int32)
    y0,y1 = clip([y-wid,y+wid],0,Data.shape[0]).astype(Int32)
    xp = x-x0
    yp = y-y0
    D = (Data[y0:y1,x0:x1]).astype(Float32)
    Good = (Mask[y0:y1,x0:x1]).astype(Float32)
    print "inter=",inter
    if sum(1.0*not_equal(D.flat,0)) > 0.9*multiply.reduce(D.shape):
       print "Doing #%d (%d,%d)" % (objid,x,y)
       B = D-getpct(D.flat,5)
       O = Outline(B,x-x0,y-y0,ring,span)
       if seeing:
          W = divz(Good,getsqrt(D))
          M = GetObjMoments(O)
#          qdump("s.fits",D)
          parameters,c = FitDonut(D,M,W,ring,span)
          Data[y0:y1,x0:x1] = (Data[y0:y1,x0:x1] - c).astype(Float32)
       else:
          Good = VTKErode(Good,3,numret=1)
          W = divz(Good,getsqrt(D))
          E = greater(VTKNonMaxSuppress(O,numret=1,usegrad=1),0.0)
          E = VTKDilate(E,5,numret=1)*D
          Y,X = indices(E.shape,Float32)
          p10 = getpct(D.flat,10)
          p38 = getpct(D.flat,38)
          sky = less(D,(p38-p10)*1 + p38)
          Xu,Yu,Du = compress(W.flat*sky.flat,[X.flat,Y.flat,D.flat],1)
          if len(Xu) < 20: return inter
          bases = transpose([power(Xu,k)*power(Yu,l) for k in range(2) for l in range(2-k)])
          bgpars = svdfit(bases,Du)
          backf = dot(bases,bgpars)
          bases = transpose([power(X.flat,k)*power(Y.flat,l) for k in range(2) for l in range(2-k)])
          backg = reshape(dot(bases,bgpars),D.shape)
          Ds = D - backg
          scat = 1.49*median(abs(backf-Du))
          Ws = greater(Ds,-3*scat)
          smo1 = VTKGauss(1.0*Ds*Ws,span/1.5/2.35,1)
          smo2 = VTKGauss(1.0*Ws,span/1.5/2.35,1)
          smo = divz(smo1,smo2)
          smo = VTKHybridMedian(smo,1)
          lmo = VTKLaplace(smo,1)
          hmo = VTKHybridMedian(lmo,1)
          Dg = divz(smo**2,hmo)
          Dg = VTKGradMag(1.0*less(Dg,0.0),1)
          Dg = 1.0*greater(Dg,0)
          Dg = Dg*greater(Ds,-2*scat)
          Dc = 1.0*Dg
#          Dg = 0.0*Dg
#          cth = 2*int(2*pi*(ring-0.5*span)+1)
#          Dg = VTKIslandRemoval(Dg,1.0,0.0,cth,numret=1)
          points = []
          Dgs = []
          while sum(Dc.flat):
             if verb: print sum(Dc.flat),;sys.stdout.flush()
             contours = VTKContours(Dc,[1])
             connect = VTKConnectivity(contours,xp,yp)
             contour = VTKArrayToNum(connect.GetPoints().GetData())
             points.append(contour[:2])
             Dg1 = VTKSeedConnectivity(Dc,[contour[0][0]],[contour[1][0]],0)
             Dgs.append(Dg1)
             Dc = Dc - Dg1
          if inter:
             qdump("s.fits",Ds*Good)
          if verb and len(points): print
          if len(points) < 2: return inter
          npoints = [point.shape[1] for point in points]
#          inner = 0
#          outer = argmax(npoints)
#          if points[inner] == points[outer]: return inter
          points = [points[jpoints] for jpoints in argsort(npoints) if npoints[jpoints] > 2*pi*(ring-span/2)]
          if points == []: return inter
          npoints = [point.shape[1] for point in points]
          x1points = array([mean(point[0]) for point in points])
          y1points = array([mean(point[1]) for point in points])
          x2points = array([sqrt(mean((point[0]-x1points[j])**2)) for j,point in enumerate(points)])
          y2points = array([sqrt(mean((point[1]-y1points[j])**2)) for j,point in enumerate(points)])
          if verb:
             print x2points
             print y2points
          keep = less(abs(x1points-xp), 2*x2points) * less(abs(y1points-yp), 2*x2points)
          keep = keep * greater(x2points,0) * greater(y2points,0)
          points = [points[j] for j in nonzero(keep)]
          if len(points) < 2: return inter
          npoints = [npoints[j] for j in nonzero(keep)]
          if verb: print npoints
          M = transpose([ [sys.stdout.write("(%d)" % (P1.shape[1])),FitEllipse(p=P1)][1] for P1 in points])
          sys.stdout.write("\n")
          mx,my,f,r,theta,a,b = M
          ds = sqrt((mx-xp)**2 + (my-yp)**2)
          order = argsort(ds)[:2]
          M = take(M,order,1)
          mx,my,f,r,theta,a,b = M
          order = argsort(a*b)[:2]
          M = take(M,order,1)
          mx,my,f,r,theta,a,b = M
          if verb:
             for j in range(2): print mx[j],my[j],a[j],b[j],r[j]
             print
          L = [DrawEllipse(X-mx,Y-my,a,b,theta,outline=1) for (mx,my,f,r,theta,a,b) in transpose(M)]
          mx,my,f,r,theta,a,b = M

          parameters = (mx[0],my[0],a[0],b[0],theta[0]*pi/180,\
                        mx[1],my[1],a[1],b[1],theta[1]*pi/180,0.0)
          Show[y0:y1,x0:x1] = (Data[y0:y1,x0:x1] + sum(L)*1e9).astype(Float32)
          if inter:
             qdump("s.fits",Ds*Good)
             qdump("g.fits",Dg*1e9)
             qdump("c.fits",sum(L)*1e9)
             proceed = raw_input("... Proceed? ")
             if proceed in ["exit","quit","done"]: os._exit(1)
             elif proceed in ["go","run"]: inter = 0
#       qdump("c.fits",c)
       if output:
          x1,y1,a1,b1,t1,x2,y2,a2,b2,t2,lseeing = parameters
          x1 += x0
          y1 += y0
          x2 += x0
          y2 += y0
          if seeing:
             a1,b1,a2,b2 = power(10, [a1,b1,a2,b2])
             p1, p2 = arctan(t1), arctan(t2)
             seesigma = pow(10,lseeing)
          else:
             p1, p2 = t1, t2
             seesigma = lseeing
          sys.stdout.write( "%4d %8.2f %8.2f %6.2f %6.2f %8.3f" % (objid,x1,y1,a1,b1,p1*180/pi)),
          sys.stdout.write( "%8.2f %8.2f %6.2f %6.2f %8.3f" % (x2,y2,a2,b2,p2*180/pi)),
          sys.stdout.write( "%10.2f %6.3f\n" % ( z,seesigma))
          sys.stdout.flush()
          output.write( "%4d %8.2f %8.2f %6.2f %6.2f %8.3f" % (objid,x1,y1,a1,b1,p1*180/pi)),
          output.write( "%8.2f %8.2f %6.2f %6.2f %8.3f" % (x2,y2,a2,b2,p2*180/pi)),
          output.write( "%10.2f %6.3f\n" % ( z,seesigma))
          output.flush()
#          exit()
       return inter

def Outline(D,x,y,ring,span,sm=1.5):
    S = VTKGauss(D,sm,1)
    Y,X = indices(S.shape,Float32)
    X,Y = X-x,Y-y
    R = sqrt(X*X+Y*Y)
    m = getpct(compress(less(R.flat,ring+span),S.flat),85)/2.0
    u = greater(S.flat,m).astype(Float32)
    i = 0
    a = 1e9
    while (abs(sum(u.flat*S.flat)-a)/a > 0.01 or not i)  and i < 10:
       i += 1
       a = sum(u.flat*S.flat)
#       print a,
       m = getpct(compress(u.flat,S.flat),50)/2.0
       u = greater(S,m).astype(Float32)
       Su = S*greater(VTKGauss(u,sm,1),0.01)
#       print sum(u.flat*S.flat)
    return u

def GetObjMoments(i,w=1.0):
    y,x = indices(i.shape)
    iw = i*w
    mx = sum(iw.flat*x.flat)/sum(iw.flat)
    my = sum(iw.flat*y.flat)/sum(iw.flat)
    mx2 = sum(iw.flat*power(x.flat,2))/sum(iw.flat) - pow(mx,2)
    my2 = sum(iw.flat*power(y.flat,2))/sum(iw.flat) - pow(my,2)
    mxy = sum(iw.flat*x.flat*y.flat)/sum(iw.flat) - mx*my
    left = (mx2 + my2)/2
    right = pow((mx2 - my2)/2,2) + pow(mxy,2)
    a2 = left + sqrt(right)
    b2 = left - sqrt(right)
    a = sqrt(2*a2)
    b = sqrt(2*b2)
    r = sqrt(a*b)
    f = sum(compress(iw.flat,i.flat))
    if mx2-my2:
       tan2t = 2 * mxy / (mx2 - my2)
       theta = arctan(tan2t)/2 * (180/pi)
    else: theta = 0.0
    return mx,my,f,r,theta,a,b

# Andrew W. Fitzgibbon, Maurizio Pilu, and Robert B. Fisher
# Direct least-squares fitting of ellipses,
# IEEE Transactions on Pattern Analysis and Machine Intelligence, 21(5), 476--480, May 1999
def FitEllipse(image=None,w=1.0,p=None):
    if image:
       mx,my,f,r,theta,sx,sy = GetObjMoments(1.0*greater(w*image,0.0))
#       qdump("g.fits",image*w)
#       print mx,my,f,r,theta,sx,sy
       W = (w*image)
       Y,X = indices(image.shape,Float32)
       Xt,Yt,Wt = compress(W.flat,[X.flat,Y.flat,W.flat])
    elif p:
       Xt,Yt = p
       Wt = ones(Xt.shape,Float32)
    f = sum(Wt)
    mx = sum(Wt*Xt)/f
    my = sum(Wt*Yt)/f
#    print 2*sum(power(Xt-mx,2)*Wt)/f
#    print 2*sum(power(Yt-my,2)*Wt)/f
    sx = sqrt(2*sum(power(Xt-mx,2)*Wt)/f)
    sy = sqrt(2*sum(power(Yt-my,2)*Wt)/f)
    print f,mx,my,sx,sy
    Xr = (Xt-mx)/sx
    Yr = (Yt-my)/sy
    D = transpose([Xr*Xr, Xr*Yr, Yr*Yr, Xr, Yr, ones(Xr.shape)])
    mm = lambda a1,a2: matrixmultiply(a1,a2)
    D = D*Wt[::,NewAxis]
    Dt = transpose(D)
    S = mm(Dt,D)
    C = zeros((6,6),Float64)
    C[0,2] = -2.0
    C[1,1] = 1.0
    C[2,0] = -2.0
    tmpA = S[:3,:3]
    tmpB = S[:3,3:]
    tmpC = S[3:,3:]
    tmpD = C[:3,:3]
    if not determinant(tmpC): return -99,-99,-99,-99,-99,-99,-99
    tmpE = mm(inverse(tmpC),transpose(tmpB))
    eigval,eigvec = eigenvectors(mm(inverse(tmpD),(tmpA-mm(tmpB,tmpE))))
    print eigval
#    print eigvec
    neg = nonzero(less(eigval.real,0))
    print neg
    if not len(neg): return -99,-99,-99,-99,-99,-99,-99
    ie = neg[0]
    A = eigvec[ie].real
    A = concatenate([A,mm(-tmpE,A)])
    print A
#    print
    par = asarray([ A[0]*sy*sy,
          A[1]*sx*sy,
          A[2]*sx*sx,
          -2*A[0]*sy*sy*mx - A[1]*sx*sy*my + A[3]*sx*sy*sy,
          -A[1]*sx*sy*mx - 2*A[2]*sx*sx*my + A[4]*sx*sx*sy,
          A[0]*sy*sy*mx*mx + A[1]*sx*sy*mx*my + A[2]*sx*sx*my*my
          - A[3]*sx*sy*sy*mx - A[4]*sx*sx*sy*my
          + A[5]*sx*sx*sy*sy ])
    print par
#    print
#    if par.typecode()!= "D": return -99,-99,-99,-99,-99,-99,-99
    theta = 0.5*arctan2(par[1],par[0]-par[2])
    cost = cos(theta); sint = sin(theta); sint2 = sint**2; cost2 = cost**2; sincost = sint*cost
    Ao = par[5]
    Au = par[3]*cost + par[4]*sint
    Av =-par[3]*sint + par[4]*cost
    Auu = par[0]*cost2 + par[2]*sint2 + par[1]*sincost
    Avv = par[0]*sint2 + par[2]*cost2 - par[1]*sincost
    if not Auu or not Avv: return -99,-99,-99,-99,-99,-99,-99
    tu = -Au/(2*Auu)
    tv = -Av/(2*Avv)
    wc = Ao - Auu*tu**2 - Avv*tv**2
    u = tu*cost - tv*sint
    v = tu*sint + tv*cost
    Ru = -wc/Auu
    Rv = -wc/Avv
#    print wc,Auu,Avv
#    print Ru,Rv
    Ru = sqrt(abs(Ru))*sign(Ru)
    Rv = sqrt(abs(Rv))*sign(Rv)
    mx,my,f,r,theta,a,b = u,v,f,sqrt(Ru*Rv),theta*180/pi,Ru,Rv
    if b > a: a,b,theta = b,a,theta+90.0

#    print mx,my,f,r,theta,a,b
#    print
#    L = DrawEllipse(X-mx,Y-my,a,b,theta,outline=1)
#    qdump("g.fits", w*image+L*1e9)
#    exit()
    return mx,my,f,r,theta,a,b

def DrawEllipse(X,Y,a,b,t,outline=0):
    t = t*pi/180
    Xt, Yt = X*cos(t)+Y*sin(t), -X*sin(t)+Y*cos(t)
    R = sqrt((Xt/a)**2 + (Yt/b)**2)
    dR = sqrt((outline/a)**2 + (outline/b)**2)
    if outline>0: E = between(R,1-dR/2.0,1+dR/2.0)
    else: E = less_equal(R,1+dR/2.0)
    return E

def getstats(s,fit=0,verb=0):
   s = sort(compress(s,s))
   mode = 0.0
   med = median(s,sorted=1)
   avg = mean(s)
   scat = 1.49*median(abs(s-med))
   lscat = log10(scat)
   bin = 1000.0
   for j in range(2):
     cen = (not j and med) or (mode)
     lo = cen - 3*scat
     hi = cen + 3*scat
     dx = (hi-lo)/bin
     x = arange(lo,hi,dx,Float32)
     n = searchsorted(s,x)
     h = diff(n)
     y = diff(x)/2.0 + x[:-1]
     w = power(h,2)
     mode = divz(sum(w*y),sum(w))
   if fit:
      m = leastsq(fitgaussian,[mode,lscat],args=(y,h,1,0),ftol=1e-4,xtol=1e-4,gtol=1e-4,maxfev=500,epsfcn=1e-4)
      if verb: print m
      mode,scat = m[0]
   scat = sqrt(mean(power(s-mode,2)))
   return mode,scat

intup = lambda x,b: pow(2,int(log(x)/log(b))+1)

def Bsky(Dr,nx,ny,nkx,nky,kx=3,ky=3):
   naxis2,naxis1 = Dr.shape
   naxis2_2 = intup(naxis2-1,2)
   naxis1_2 = intup(naxis1-1,2)
   D = zeros((naxis2_2,naxis1_2),Dr.typecode())
   D[:Dr.shape[0],:Dr.shape[1]] = Dr
   nx = int(1.0*naxis1_2/naxis1*nx+1)
   ny = int(1.0*naxis2_2/naxis2*ny+1)
   print "Scaling to (nx,ny) = (%d,%d)" % (nx,ny)
   print "Using (nx,ny) = (%d,%d)" % (nx,ny)
   mx = int(1.0*naxis1_2/nx)
   my = int(1.0*naxis2_2/ny)
   good = ones(D.shape,Float32)
   print "Computing binned image statistics."
   Dmode,Dscat = ShrinkN(D,good,mx,my,func=lambda s: getstats(s,fit=0,verb=0),nret=2)
   Dgood = (Dmode > 0.0)*(Dscat > 0.0)
   Dy,Dx = indices(Dmode.shape,Float32)
   z,w,y,x, = compress(Dgood.flat,[Dmode.flat,Dscat.flat,Dy.flat,Dx.flat],1)
   w = 1/w
   xmin,xmax = extrema(x)
   ymin,ymax = extrema(y)
   dx = (xmax-xmin)/(nkx+1)
   dy = (ymax-ymin)/(nky+1)
   tx = arange(xmin,xmax+dx/2.0,dx,Float32)
   ty = arange(ymin,ymax+dy/2.0,dy,Float32)
   while len(tx) < 2*kx+2: tx = concatenate([[tx[0]],tx,[tx[-1]]])
   while len(ty) < 2*ky+2: ty = concatenate([[ty[0]],ty,[ty[-1]]])
   tck = bisplrep(x,y,z,w=w,s=0,task=-1,kx=kx,ky=ky,tx=tx,ty=ty,quite=1,\
        xb=-0.5,xe=Dmode.shape[1]-0.5,yb=-0.5,ye=Dmode.shape[0]-0.5)
   ox = arange(0,Dmode.shape[1],1,Float32)
   oy = arange(0,Dmode.shape[0],1,Float32)
   Dspline = 1*transpose(bisplev(ox,oy,tck))
   ox = arange(-0.5,Dmode.shape[1],1.0/mx,Float32)
   oy = arange(-0.5,Dmode.shape[0],1.0/my,Float32)
   Dfit = 1*transpose(bisplev(ox,oy,tck))
   Dfit = Dfit[:naxis2,:naxis1].astype(Float32)
   return Dfit


